#  Copyright 2023 Google LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
import sys
from typing import List, Dict, Any, Union

from googleapiclient import discovery

from gcp_scanner.crawler.interface_crawler import ICrawler


class ComputeSubnetsCrawler(ICrawler):
  """Handle crawling of compute subnets data."""

  def crawl(self, project_name: str, service: discovery.Resource,
            config: Dict[str, Union[bool, str]] = None) -> List[Dict[str, Any]]:
    """Retrieve a list of subnets available in the project.

     Args:
         project_name: The name of the project to query information about.
         service: A resource object for interacting with the GCP API.
         config: Configuration options for the crawler (Optional).

     Returns:
         A list of resource objects representing the crawled data.
     """
    logging.info("Retrieving Subnets")
    subnets_list = list()
    try:
      request = service.subnetworks().aggregatedList(project=project_name)
      while request is not None:
        response = request.execute()
        if response.get("items", None) is not None:
          subnets_list.extend(list(response["items"].items()))
        request = service.subnetworks().aggregatedList_next(
          previous_request=request, previous_response=response)
    except Exception:
      logging.info("Failed to get subnets in the %s", project_name)
      logging.info(sys.exc_info())

    return subnets_list



#  Copyright 2023 Google LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import sys
from typing import List, Dict, Any, Union

from googleapiclient import discovery

from gcp_scanner.crawler.interface_crawler import ICrawler


class KMSKeysCrawler(ICrawler):
  '''Handle crawling of KMS Keys data.'''

  def crawl(self, project_id: str, service: discovery.Resource,
            config: Dict[str, Union[bool, str]] = None) -> List[Dict[str, Any]]:
    '''Retrieve a list of KMS Keys available in the project.

    Args:
      project_id: A name of a project to query info about.
      service: A resource object for interacting with KMS API.
      config: Configuration options for the crawler (Optional).

    Returns:
      A list of resource objects representing the crawled data.
    '''

    logging.info("Retrieving KMS keys")
    kms_keys_list = list()
    try:
      # list all possible locations
      locations_list = list()
      request = service.projects().locations().list(name=f"projects/{project_id}")
      while request is not None:
        response = request.execute()
        for location in response.get("locations", []):
          locations_list.append(location["locationId"])
        request = service.projects().locations().list_next(
          previous_request=request, previous_response=response)

      for location_id in locations_list:
        request_loc = service.projects().locations().keyRings().list(
          parent=f"projects/{project_id}/locations/{location_id}")
        while request_loc is not None:
          response_loc = request_loc.execute()
          for keyring in response_loc.get("keyRings", []):
            request = service.projects().locations().keyRings().cryptoKeys().list(
              parent=keyring["name"])
            while request is not None:
              response = request.execute()
              for key in response.get("cryptoKeys", []):
                kms_keys_list.append(key)

              request = service.projects().locations().keyRings().cryptoKeys(
              ).list_next(
                previous_request=request, previous_response=response)

          request_loc = service.projects().locations().keyRings().list_next(
            previous_request=request, previous_response=response)
    except Exception:
      logging.info("Failed to retrieve KMS keys for project %s", project_id)
      logging.info(sys.exc_info())
    return kms_keys_list



#  Copyright 2023 Google LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from googleapiclient import discovery
from httplib2 import Credentials

from .interface_client import IClient


class DatastoreClient(IClient):
  """DatastoreClient class."""

  def get_service(self, credentials: Credentials) -> discovery.Resource:
    """Get discovery service for datastore resource.

    Args:
      credentials: An google.oauth2.credentials.Credentials object.

    Returns:
      An object of discovery.Resource
    """
    return discovery.build(
      "datastore",
      "v1",
      credentials=credentials,
      cache_discovery=False,
    )



#  Copyright 2023 Google LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import sys
from typing import List, Dict, Any, Union

from googleapiclient import discovery

from gcp_scanner.crawler.interface_crawler import ICrawler


class FirestoreCollectionsCrawler(ICrawler):
  """Handle crawling of Firestore data."""

  def crawl(self, project_id: str, service: discovery.Resource,
            config: Dict[str, Union[bool, str]] = None) -> Dict[str, List[Dict[str, Any]]]:
    """Retrieve a list of Firestore datasets available in the project.

    Args:
      project_id: A name of a project to query info about.
      service: A resource object for interacting with the Firestore API.
      config: Configuration options for the crawler (Optional).

    Returns:
      A list of resource objects representing the crawled data.
    """

    logging.info("Retrieving Firestore Databases")
    firestore_documents = dict()
    project_path = f"projects/{project_id}"
    try:
      request = service.projects().databases().list(parent=project_path)
      if request is not None:
        response = request.execute()
        for database in response.get('databases', []):
          db_name = database['name']
          firestore_documents[db_name] = self.get_firestore_collectionIds(
            db_name,
            service,
          )
    except Exception:
      logging.info("Failed to retrieve Firestore databases for project %s", project_id)
      logging.info(sys.exc_info())
    return firestore_documents

  def get_firestore_collectionIds(self, parent: str, service: discovery.Resource) -> List[Dict[str, Any]]:
    """Retrieve a list of Firestore collections available in the dataset.

    Args:
      parent: string, Required. The parent document. In the format: `projects/{project_id}/databases/{database_id}/documents/{document_path}`.
      service: A resource object for interacting with the Firestore API.

    Returns:
      A list of Firestore collection ids.
    """

    logging.info("Firestore documents for %s", parent)
    list_of_collections_ids = list()
    try:
      request = service.projects().databases().documents().listCollectionIds(
        parent=f"{parent}/documents/*/**"
      )
      request.uri = request.uri.replace('/*/**', '')
      while request is not None:
        response = request.execute()
        list_of_collections_ids.extend(response.get("collectionIds", None))
        request = service.projects().databases().documents().listCollectionIds_next(
          previous_request=request,
          previous_response=response,
        )
    except Exception as ex:
      logging.info("Failed to retrieve Firestore collections for %s", parent)
      logging.info(sys.exc_info())
    return list_of_collections_ids



"""
The entry point for the visualization tool.

This module is responsible for starting the HTTP server and serving the
static files for the GUI.
"""

import argparse
import http.server
import socketserver
import os


def main():
  parser = argparse.ArgumentParser(
      description='GCP Scanner GUI',
      usage='python3 %(prog)s -p 8080')
  parser.add_argument(
      '-p',
      '--port',
      default=8080,
      type=int,
      dest='port',
      help='Port to serve the GUI on')
  args = parser.parse_args()

  # change directory to the GUI directory
  os.chdir(os.path.dirname(os.path.abspath(__file__)) + '/static')

  handler = http.server.SimpleHTTPRequestHandler
  with socketserver.TCPServer(('', args.port), handler) as httpd:
    message = 'Running at: http://localhost:' + str(args.port)
    print('\033[92m' + message + '\033[0m')
    httpd.serve_forever()


if __name__ == '__main__':
  main()



#  Copyright 2023 Google LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging

from gcp_scanner.client.appengine_client import AppEngineClient
from gcp_scanner.client.bigquery_client import BQClient
from gcp_scanner.client.bigtable_client import BigTableClient
from gcp_scanner.client.cloud_billing_client import CloudBillingClient
from gcp_scanner.client.cloud_functions_client import CloudFunctionsClient
from gcp_scanner.client.cloud_resource_manager_client import CloudResourceManagerClient
from gcp_scanner.client.compute_client import ComputeClient
from gcp_scanner.client.datastore_client import DatastoreClient
from gcp_scanner.client.dns_client import DNSClient
from gcp_scanner.client.domains_client import DomainsClient
from gcp_scanner.client.filestore_client import FilestoreClient
from gcp_scanner.client.firestore_client import FirestoreClient
from gcp_scanner.client.iam_client import IAMClient
from gcp_scanner.client.kms_client import CloudKMSClient
from gcp_scanner.client.pubsub_client import PubSubClient
from gcp_scanner.client.service_management_client import ServiceManagementClient
from gcp_scanner.client.serviceusage_client import ServiceUsageClient
from gcp_scanner.client.sourcerepo_client import SourceRepoClient
from gcp_scanner.client.spanner_client import SpannerClient
from gcp_scanner.client.sql_client import SQLClient
from gcp_scanner.client.storage_client import StorageClient


class ClientFactory:
  """Factory class for creating clients."""

  clients = {
    "appengine": AppEngineClient,
    "bigquery": BQClient,
    "bigtableadmin": BigTableClient,
    "cloudbilling": CloudBillingClient,
    "cloudfunctions": CloudFunctionsClient,
    "cloudkms": CloudKMSClient,
    "cloudresourcemanager": CloudResourceManagerClient,
    "compute": ComputeClient,
    "datastore": DatastoreClient,
    "domains": DomainsClient,
    "dns": DNSClient,
    "firestore": FirestoreClient,
    "file": FilestoreClient,
    "iam": IAMClient,
    "pubsub": PubSubClient,
    "servicemanagement": ServiceManagementClient,
    "serviceusage": ServiceUsageClient,
    "sourcerepo": SourceRepoClient,
    "spanner": SpannerClient,
    "sqladmin": SQLClient,
    "storage": StorageClient,
  }

  @classmethod
  def get_client(cls, name):
    """Returns the appropriate client."""

    client_cls = cls.clients.get(name.lower())
    if client_cls:
      return client_cls()

    logging.error("Client not supported.")
    return None



#  Copyright 2023 Google LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging

from gcp_scanner.crawler.app_services_crawler import AppServicesCrawler
from gcp_scanner.crawler.bigquery_crawler import BigQueryCrawler
from gcp_scanner.crawler.bigtable_instances_crawler import BigTableInstancesCrawler
from gcp_scanner.crawler.cloud_billing_account_crawler import CloudBillingAccountCrawler
from gcp_scanner.crawler.cloud_functions_crawler import CloudFunctionsCrawler
from gcp_scanner.crawler.cloud_resource_manager_iam_policy_crawler import CloudResourceManagerIAMPolicyCrawler
from gcp_scanner.crawler.cloud_resource_manager_project_info_crawler import CloudResourceManagerProjectInfoCrawler
from gcp_scanner.crawler.cloud_resource_manager_project_list_crawler import CloudResourceManagerProjectListCrawler
from gcp_scanner.crawler.compute_disks_crawler import ComputeDisksCrawler
from gcp_scanner.crawler.compute_firewall_rules_crawler import ComputeFirewallRulesCrawler
from gcp_scanner.crawler.compute_images_crawler import ComputeImagesCrawler
from gcp_scanner.crawler.compute_instances_crawler import ComputeInstancesCrawler
from gcp_scanner.crawler.compute_security_policies_crawler import ComputeSecurityPoliciesCrawler
from gcp_scanner.crawler.compute_snapshots_crawler import ComputeSnapshotsCrawler
from gcp_scanner.crawler.compute_static_ips_crawler import ComputeStaticIPsCrawler
from gcp_scanner.crawler.compute_subnets_crawler import ComputeSubnetsCrawler
from gcp_scanner.crawler.datastore_crawler import DatastoreCrawler
from gcp_scanner.crawler.dns_managed_zones_crawler import DNSManagedZonesCrawler
from gcp_scanner.crawler.dns_policies_crawler import DNSPoliciesCrawler
from gcp_scanner.crawler.domains_crawler import DomainsCrawler
from gcp_scanner.crawler.endpoints_crawler import EndpointsCrawler
from gcp_scanner.crawler.filestore_instances_crawler import FilestoreInstancesCrawler
from gcp_scanner.crawler.firestore_collections_crawler import FirestoreCollectionsCrawler
from gcp_scanner.crawler.kms_keys_crawler import KMSKeysCrawler
from gcp_scanner.crawler.machine_images_crawler import ComputeMachineImagesCrawler
from gcp_scanner.crawler.pubsub_subscriptions_crawler import PubSubSubscriptionsCrawler
from gcp_scanner.crawler.service_accounts_crawler import ServiceAccountsCrawler
from gcp_scanner.crawler.service_usage_crawler import ServiceUsageCrawler
from gcp_scanner.crawler.source_repo_crawler import CloudSourceRepoCrawler
from gcp_scanner.crawler.spanner_instances_crawler import SpannerInstancesCrawler
from gcp_scanner.crawler.sql_instances_crawler import SQLInstancesCrawler
from gcp_scanner.crawler.storage_buckets_crawler import StorageBucketsCrawler

service_crawler_map = {
  "app_services": AppServicesCrawler,
  "bigtable_instances": BigTableInstancesCrawler,
  "bq": BigQueryCrawler,
  "cloud_billing_account": CloudBillingAccountCrawler,
  "cloud_functions": CloudFunctionsCrawler,
  "compute_disks": ComputeDisksCrawler,
  "compute_images": ComputeImagesCrawler,
  "compute_instances": ComputeInstancesCrawler,
  "compute_security_policies": ComputeSecurityPoliciesCrawler,
  "compute_snapshots": ComputeSnapshotsCrawler,
  "datastore_kinds": DatastoreCrawler,
  "dns_policies": DNSPoliciesCrawler,
  "endpoints": EndpointsCrawler,
  "filestore_instances": FilestoreInstancesCrawler,
  "firestore_collections": FirestoreCollectionsCrawler,
  "firewall_rules": ComputeFirewallRulesCrawler,
  "iam_policy": CloudResourceManagerIAMPolicyCrawler,
  "kms": KMSKeysCrawler,
  "machine_images": ComputeMachineImagesCrawler,
  "managed_zones": DNSManagedZonesCrawler,
  "project_info": CloudResourceManagerProjectInfoCrawler,
  "project_list": CloudResourceManagerProjectListCrawler,
  "pubsub_subs": PubSubSubscriptionsCrawler,
  "registered_domains": DomainsCrawler,
  "services": ServiceUsageCrawler,
  "service_accounts": ServiceAccountsCrawler,
  "sourcerepos": CloudSourceRepoCrawler,
  "spanner_instances": SpannerInstancesCrawler,
  "sql_instances": SQLInstancesCrawler,
  "static_ips": ComputeStaticIPsCrawler,
  "storage_buckets": StorageBucketsCrawler,
  "subnets": ComputeSubnetsCrawler,
}


class CrawlerFactory:
  """Factory class for creating crawlers."""

  @classmethod
  def create_crawler(cls, name):
    """Returns the appropriate crawler."""
    if name in service_crawler_map:
      return service_crawler_map[name]()

    logging.error("Crawler not supported.")
    return None



#  Copyright 2023 Google LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import sys
from typing import Dict, Any, Union

from googleapiclient import discovery

from gcp_scanner.crawler.interface_crawler import ICrawler


class CloudSourceRepoCrawler(ICrawler):
  '''Handle crawling of Cloud Source Repo data.'''

  def crawl(self, project_id: str, service: discovery.Resource,
            config: Dict[str, Union[bool, str]] = None) -> Dict[str, Any]:
    """Retrieve a list of cloud source repositories enabled in the project.

    Args:
    project_id: An id of a project to query info about.
    service: A resource object for interacting with the Source Repo API.
    config: Configuration options for the crawler (Optional).

    Returns:
    A list of cloud source repositories in the project.
    """

    logging.info("Retrieving cloud source repositories %s", project_id)
    list_of_repos = list()

    request = service.projects().repos().list(
      name="projects/" + project_id,
      pageSize=500
    )
    try:
      while request is not None:
        response = request.execute()
        list_of_repos.extend(response.get("repos", None))

        request = service.projects().repos().list_next(
          previous_request=request,
          previous_response=response
        )
    except Exception:
      logging.info("Failed to retrieve source repos for project %s", project_id)
      logging.info(sys.exc_info())
    return list_of_repos



#  Copyright 2023 Google LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import sys
from typing import List, Dict, Any, Union

from googleapiclient import discovery

from gcp_scanner.crawler.interface_crawler import ICrawler


class DatastoreCrawler(ICrawler):
  """Handle crawling of cloud datastore data."""

  def crawl(self, project_id: str, service: discovery.Resource,
            config: Dict[str, Union[bool, str]] = None) -> Dict[str, List[Dict[str, Any]]]:
    """Retrieve a list of datastore kinds available in the project.

    Args:
      project_id: A name of a project to query info about.
      service: A resource object for interacting with the Firestore API.
      config: Configuration options for the crawler (Optional).

    Returns:
      A list of objects representing the crawled data.
    """

    logging.info("Retrieving Datastore entities.")
    datastore_kinds = dict()
    query = {
      "query": {
        "kind": [{"name": "__kind__"}]
      }
    }
    try:
      request = service.projects().runQuery(projectId=project_id, body=query)
      if request is not None:
        response = request.execute()
        entity_results = response.get("batch", {}).get("entityResults", [])
        datastore_kinds['kinds'] = [entity["entity"]["key"]["path"][0]["name"] for entity in entity_results]
    except Exception:
      logging.info("Failed to retrieve datastore entities for project %s", project_id)
      logging.info(sys.exc_info())
    return datastore_kinds



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""The module with SpiderContext query model.

"""

import queue
from typing import List, Tuple

from httplib2 import Credentials


class SpiderContext:
  """A simple class to initialize the context with a list of root SAs
  """

  def __init__(self, sa_tuples: List[Tuple[str, Credentials, List[str]]]):
    """Initialize the context with a list of the root service accounts.

    Args:
      sa_tuples: [(sa_name, sa_object, chain_so_far)]
    """

    self.service_account_queue = queue.Queue()
    for sa_tuple in sa_tuples:
      self.service_account_queue.put(sa_tuple)

  def __repr__(self) -> str:
    return f"{list(self.service_account_queue.queue)}"


class ProjectInfo:
  """A simple class to store project scan configration.
  """

  def __init__(
    self,
    project,
    sa_results,
    out_dir,
    scan_config,
    light_scan,
    target_project,
    scan_time_suffix,
    sa_name,
    credentials,
    chain_so_far,
    resource_worker_count
  ):
    self.project = project
    self.sa_results = sa_results
    self.out_dir = out_dir
    self.scan_config = scan_config
    self.light_scan = light_scan
    self.target_project = target_project
    self.scan_time_suffix = scan_time_suffix
    self.sa_name = sa_name
    self.credentials = credentials
    self.chain_so_far = chain_so_far
    self.resource_worker_count = resource_worker_count



#  Copyright 2023 Google LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
import logging
import sys
from typing import List, Dict, Any, Union, Tuple, TextIO, Optional

from googleapiclient import discovery, errors

from gcp_scanner.crawler.interface_crawler import ICrawler


class StorageBucketsCrawler(ICrawler):
  """Handle crawling of bucket names data."""

  _config_dependency = True # Define that config file is needed

  def crawl(self, project_name: str, service: discovery.Resource,
            config: Dict[str, Union[bool, str]] = None) -> Dict[str, Tuple[Any, List[Any]]]:
    """Retrieve a list of buckets available in the project.

     Args:
         project_name: The name of the project to query information about.
         service: A resource object for interacting with the GCP API.
         config: Configuration options for the crawler (Optional).

     Returns:
         A list of resource objects representing the crawled data.
     """
    logging.info("Retrieving GCS Buckets")
    # prepare additional configs
    dump_fd = self._get_dump_file_dir(config=config)
    is_dump_iam_policies = self._get_is_dump_iam_policies(config=config)
    # output dict
    buckets_dict = dict()
    # Make an authenticated API request
    request = service.buckets().list(project=project_name)
    while request is not None:
      try:
        response = request.execute()
      except Exception:
        logging.info("Failed to list buckets in the %s", project_name)
        logging.info(sys.exc_info())
        break

      for bucket in response.get("items", []):
        buckets_dict[bucket["name"]] = bucket
        if is_dump_iam_policies is True:
          buckets_dict[bucket["name"]]["iam_policy"] = self._get_bucket_iam(bucket["name"], service)
        if dump_fd is not None:
          ret_fields = "nextPageToken,items(bucket,name,size,contentType, timeCreated)"
          req = service.objects().list(bucket=bucket["name"], fields=ret_fields)

          while req:
            try:
              resp = req.execute()
              for item in resp.get("items", []):
                dump_fd.write(json.dumps(item, indent=2, sort_keys=False))

              req = service.objects().list_next(req, resp)
            except errors.HttpError:
              logging.info("Failed to read the bucket %s", bucket["name"])
              logging.info(sys.exc_info())
              break

      request = service.buckets().list_next(
        previous_request=request, previous_response=response)
    if dump_fd is not None:
      dump_fd.close()
    return buckets_dict
  
  @property
  def has_config_dependency(self) -> bool:
    """Checks if the class needs a config file

    Returns:
        bool: Returns config_dependency private variable which is False by default.
    """
    return self._config_dependency

  @classmethod
  def _get_bucket_iam(cls, bucket_name: str, service: discovery.Resource) -> List[Any]:
    """Retrieve IAM policies in the bucket.

    Args:
      bucket_name: A name of bucket to query info about.
      service: An authenticated API request.
    Returns:
      A list with bucket IAM policies.
    """

    logging.info("Retrieving GCS Bucket %s IAM Policy", bucket_name)
    bucket_iam_policies = list()
    request = service.buckets().getIamPolicy(bucket=bucket_name)
    try:
      response = request.execute()
    except Exception:
      logging.info("Failed to IAM Policy in the %s", bucket_name)
      logging.info(sys.exc_info())
      return []

    for bucket_iam_policy in response.get("bindings", []):
      bucket_iam_policies.append(bucket_iam_policy)

    return bucket_iam_policies

  @classmethod
  def _get_dump_file_dir(cls, config: Dict[str, Union[bool, str]]) -> Optional[TextIO]:
    """Get the dump file directory based on the provided configuration.

    Args:
        config: Configuration dictionary with keys 'fetch_file_names' (bool) and 'gcs_output_path' (str).

    Returns:
        TextIO object for the dump file if 'fetch_file_names' is True and 'gcs_output_path' is provided. Otherwise, None.
    """
    dump_file_names = None
    if config is not None and config.get('fetch_file_names', False) is True:
      gcs_output_path = config.get('gcs_output_path', '')  # think a good fallback if gcs_output_path is not set
      dump_file_names = open(gcs_output_path, 'w', encoding='utf-8')
    return dump_file_names

  @classmethod
  def _get_is_dump_iam_policies(cls, config: Dict[str, Union[bool, str]]) -> bool:
    """Check if dumping IAM policies is enabled based on the provided configuration.

    Args:
        config: Configuration dictionary with a key 'fetch_buckets_iam' (bool) indicating whether to fetch IAM policies.

    Returns:
        True if 'fetch_buckets_iam' is True in the config. Otherwise, False.
    """
    if config is not None and config.get('fetch_buckets_iam', False) is True:
      return True
    return False



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Command-Line Argument Parser

"""

import argparse
import logging
import os
import sys

from .error_handler import ERROR_CODES

def arg_parser():
  """Creates an argument parser using the `argparse` module and defines
  several command-line arguments.

  Args:
    None

  Returns:
    argparse.Namespace: A namespace object containing the parsed command-line
    arguments.
  """
  parser = argparse.ArgumentParser(
      prog='scanner.py',
      description='GCP Scanner',
      usage='python3 %(prog)s -o folder_to_save_results -g -')
  required_named = parser.add_argument_group('Required parameters')
  required_named.add_argument(
      '-o',
      '--output-dir',
      required=True,
      dest='output',
      default='scan_db',
      help='Path to output directory')
  parser.add_argument(
      '-ls',
      '--light-scan',
      default=False,
      dest='light_scan',
      action='store_true',
      help='Return only the most important GCP resource fields in the output.')
  parser.add_argument(
      '-k',
      '--sa-key-path',
      default=None,
      dest='key_path',
      help='Path to directory with SA keys in json format')
  parser.add_argument(
      '-g',
      '--gcloud-profile-path',
      default=None,
      dest='gcloud_profile_path',
      help='Path to directory with gcloud profile. Specify -\
 to search for credentials in default gcloud config path'
  )
  parser.add_argument(
      '-m',
      '--use-metadata',
      default=False,
      dest='use_metadata',
      action='store_true',
      help='Extract credentials from GCE instance metadata')
  parser.add_argument(
      '-at',
      '--access-token-files',
      default=None,
      dest='access_token_files',
      help='A list of comma separated files with access token and OAuth scopes.\
TTL limited. A token and scopes should be stored in JSON format.')
  parser.add_argument(
      '-rt',
      '--refresh-token-files',
      default=None,
      dest='refresh_token_files',
      help='A list of comma separated files with refresh_token, client_id,\
token_uri and client_secret stored in JSON format.'
  )
  parser.add_argument(
      '-s',
      '--service-account',
      default=None,
      dest='key_name',
      help='Name of individual SA to scan')
  parser.add_argument(
      '-p',
      '--project',
      default=None,
      dest='target_project',
      help='Name of individual project to scan')
  parser.add_argument(
      '-f',
      '--force-projects',
      default=None,
      dest='force_projects',
      help='Comma separated list of project names to include in the scan')
  parser.add_argument(
      '-c',
      '--config',
      default=None,
      dest='config_path',
      help='A path to config file with a set of specific resources to scan.')
  parser.add_argument(
      '-l',
      '--logging',
      default='WARNING',
      dest='log_level',
      choices=('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'),
      help='Set logging level (INFO, WARNING, ERROR)')
  parser.add_argument(
      '-lf',
      '--log-file',
      default=None,
      dest='log_file',
      help='Save logs to the path specified rather than displaying in\
 console')
  parser.add_argument(
      '-pwc',
      '--project-worker-count',
      default=1,
      dest='project_worker_count',
      help='Set limit for project crawlers run in parallel.')
  parser.add_argument(
      '-rwc',
      '--resource-worker-count',
      default=1,
      dest='resource_worker_count',
      help='Set limit for resource crawlers run in parallel.')

  args: argparse.Namespace = parser.parse_args()

  if not any(
          [args.key_path,
           args.gcloud_profile_path,
           args.use_metadata,
           args.access_token_files,
           args.refresh_token_files]):
    logging.error(
        'Please select at least one option to begin scan\
 -k/--sa-key-path,-g/--gcloud-profile-path, -m, -rt, -at'
    )
  if not os.path.isdir(args.output):
    logging.error(
      '\"%s\" doesn\'t exist. Please enter a valid directory path.', 
      args.output
    )
    sys.exit(ERROR_CODES.get('InvalidDirError'))

  return args



#  Copyright 2023 Google LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List, Union

from googleapiclient import discovery


class ICrawler(metaclass=ABCMeta):
  """Interface for Crawler Classes.

  This interface defines the contract for crawler classes that are used to crawl resource data
  of a given project using the Google Cloud Platform (GCP) API.

  """

  """Variable to identify if config file is needed

  Access Type: Private
  """
  _config_dependency = False

  @staticmethod
  @abstractmethod
  def crawl(project_name: str, service: discovery.Resource,
            config: Dict[str, Union[bool, str]] = None) -> List[Dict[str, Any]]:
    """Crawl resource data of the given project.

    This method retrieves resource data of the specified project using the provided GCP API service
    and returns a list of resource objects.

    Args:
        project_name: The name of the project to query information about.
        service: A resource object for interacting with the GCP API.
        config: Configuration options for the crawler (Optional).

    Returns:
        A list of resource objects representing the crawled data.

    Raises:
        NotImplementedError: If a child class does not implement this method.

    """

    raise NotImplementedError("Child class must implement the crawl() method.")
  
  @property
  def has_config_dependency(self) -> bool:
    """Checks if the class needs a config file

    Returns:
        bool: Returns config_dependency private variable which is False by default.
    """
    return self._config_dependency


# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""The main module that initiates scanning of GCP resources."""
import collections
from datetime import datetime
import json
from json.decoder import JSONDecodeError
import logging
import os
from pathlib import Path
import sys
import threading
import time
from typing import Any, Dict, List, Optional, Union

from google.auth.exceptions import MalformedError
from google.cloud import container_v1
from google.cloud import iam_credentials
from google.cloud.iam_credentials_v1.services.iam_credentials.client import IAMCredentialsClient
from httplib2 import Credentials

from . import arguments
from . import credsdb
from . import models
from . import scanner
from .client.client_factory import ClientFactory
from .crawler import misc_crawler
from .crawler.crawler_factory import CrawlerFactory

# We define the schema statically to make it easier for the user and avoid extra
# config files.
LIGHT_VERSION_SCAN_SCHEMA = {
    'compute_instances': [
        'name',
        'zone',
        'machineType',
        'networkInterfaces',
        'status',
    ],
    'compute_images': ['name', 'status', 'diskSizeGb', 'sourceDisk'],
    'machine_images': [
        'name',
        'description',
        'status',
        'sourceInstance',
        'totalStorageBytes',
        'savedDisks',
    ],
    'compute_disks': [
        'name',
        'sizeGb',
        'zone',
        'status',
        'sourceImage',
        'users',
    ],
    'compute_snapshots': ['name', 'status', 'sourceDisk', 'downloadBytes'],
    'managed_zones': ['name', 'dnsName', 'description', 'nameServers'],
    'sql_instances': [
        'name',
        'region',
        'ipAddresses',
        'databaseVersion',
        'state',
    ],
    'cloud_functions': [
        'name',
        'eventTrigger',
        'status',
        'entryPoint',
        'serviceAccountEmail',
    ],
    'kms': ['name', 'primary', 'purpose', 'createTime'],
    'services': ['name'],
}

# The following map is used to establish the relationship between
# crawlers and clients. It determines the appropriate crawler and
# client to be selected from the respective factory classes.
CRAWL_CLIENT_MAP = {
    'app_services': 'appengine',
    'bigtable_instances': 'bigtableadmin',
    'bq': 'bigquery',
    'cloud_billing_account': 'cloudbilling',
    'cloud_functions': 'cloudfunctions',
    'compute_disks': 'compute',
    'compute_images': 'compute',
    'compute_instances': 'compute',
    'compute_security_policies':'compute',
    'compute_snapshots': 'compute',
    'datastore_kinds': 'datastore',
    'dns_policies': 'dns',
    'endpoints': 'servicemanagement',
    'firestore_collections': 'firestore',
    'filestore_instances': 'file',
    'firewall_rules': 'compute',
    'iam_policy': 'cloudresourcemanager',
    'kms': 'cloudkms',
    'machine_images': 'compute',
    'managed_zones': 'dns',
    'pubsub_subs': 'pubsub',
    'registered_domains': 'domains',
    'services': 'serviceusage',
    'service_accounts': 'iam',
    'sourcerepos': 'sourcerepo',
    'spanner_instances': 'spanner',
    'sql_instances': 'sqladmin',
    'static_ips': 'compute',
    'storage_buckets': 'storage',
    'subnets': 'compute',
}


def is_set(config: Optional[dict], config_setting: str) -> Union[dict, bool]:
  if config is None:
    return True
  obj = config.get(config_setting, {})
  return obj.get('fetch', False)


def save_results(res_data: Dict, res_path: str, is_light: bool):
  """The function to save scan results on disk in json format.

  Args:
    res_data: scan results as a dictionary of entries
    res_path: full path to save data in file
    is_light: save only the most interesting results
  """

  if is_light is True:
    # returning the light version of the scan based on predefined schema
    for gcp_resource, schema in LIGHT_VERSION_SCAN_SCHEMA.items():
      projects = res_data.get('projects', {})
      for project_name, project_data in projects.items():
        scan_results = project_data.get(gcp_resource, {})
        light_results = list()
        for scan_result in scan_results:
          light_results.append({key: scan_result.get(key) for key in schema})

        project_data.update({gcp_resource: light_results})
        projects.update({project_name: project_data})
      res_data.update({'projects': projects})

  # Write out results to json DB
  sa_results_data = json.dumps(res_data, indent=2, sort_keys=False)

  with open(res_path, 'a', encoding='utf-8') as outfile:
    outfile.write(sa_results_data)


def get_crawl(
    crawler: Any,
    project_id: str,
    client: Any,
    crawler_config: dict,
    scan_results: dict,
    crawler_name: str,
):
  """The function calls the crawler and returns result in dictionary

  Args:
    crawler: crawler method to start
    project_id: id of a project to scan
    client: appropriate client method
    crawler_config: a dictionary containing specific parameters for a crawler
    scan_results: a dictionary to save scanning results
    crawler_name: name of a crawler

  Returns:
    scan_result: a dictionary with scanning results
  """
  if crawler.has_config_dependency:
    res = crawler.crawl(project_id, client, crawler_config)
  else:
    res = crawler.crawl(project_id, client)
  if res is not None and len(res) != 0:
    scan_results[crawler_name] = res
  return scan_results


def get_resources(project: models.ProjectInfo):
  """The function crawls the data for a project and stores the results in a

     dictionary.

  Args:
    project: class to store project scan configration
  """

  if (
      project.target_project
      and project.target_project not in project.project['projectId']
  ):
    return

  project_id = project.project['projectId']
  print(f'Inspecting project {project_id}')
  project_result = dict()

  project_result['project_info'] = project.project
  project_result['service_account_chain'] = project.sa_results[
      'service_account_chain'
  ]
  project_result['current_service_account'] = project.sa_results[
      'current_service_account'
  ]
  project_result['token_scopes'] = project.sa_results['token_scopes']
  project_result['service_account_edges'] = project.sa_results[project_id][
      'service_account_edges'
  ]

  # Fail with error if the output file already exists
  output_file_name = f'{project_id}-{project.scan_time_suffix}.json'
  output_path = Path(project.out_dir, output_file_name)
  gcs_output_path = Path(project.out_dir, f'gcs-{output_file_name}')

  try:
    with open(output_path, 'x', encoding='utf-8'):
      pass

  except FileExistsError:
    logging.error(
        'Try removing the %s file and restart the scanner.', output_file_name
    )

  threads_list = list()
  for crawler_name, client_name in CRAWL_CLIENT_MAP.items():
    if is_set(project.scan_config, crawler_name):
      crawler_config = {}
      if project.scan_config is not None:
        crawler_config = project.scan_config.get(crawler_name)

      # add gcs output path to the config.
      # this path is used by the storage bucket crawler as of now.
      crawler_config['gcs_output_path'] = gcs_output_path

      # crawl the data
      crawler = CrawlerFactory.create_crawler(crawler_name)
      client = ClientFactory.get_client(client_name).get_service(
          project.credentials,
      )

      t = threading.Thread(
          target=get_crawl,
          args=(
              crawler,
              project_id,
              client,
              crawler_config,
              project_result,
              crawler_name,
          ),
      )
      t.daemon = True
      t.start()
      threads_list.append(t)

      while True:
        active_threads = 0
        for t in threads_list:
          if t.is_alive():
            active_threads += 1
        if active_threads >= project.resource_worker_count:
          time.sleep(0.1)
        else:
          break

  for t in threads_list:
    t.join()

  # Call other miscellaneous crawlers here
  if is_set(project.scan_config, 'gke_clusters'):
    gke_client = gke_client_for_credentials(project.credentials)
    res = misc_crawler.get_gke_clusters(
        project_id,
        gke_client,
    )
    if res is not None and len(res) != 0:
      project_result['gke_clusters'] = res
  if is_set(project.scan_config, 'gke_images'):
    res = misc_crawler.get_gke_images(
        project_id,
        project.credentials.token,
    )
    if res is not None and len(res) != 0:
      project_result['gke_images'] = res

  logging.info('Saving results for %s into the file', project_id)
  save_results(project_result, output_path, project.light_scan)


def impersonate_service_accounts(
    context,
    project,
    scan_config,
    sa_results,
    chain_so_far,
    sa_name,
    credentials,
):
  """The function enumerates projects accessible by SA and impersonates them."""

  # Enumerate projects accessible by SA
  project_id = project['projectId']
  project_result = sa_results['projects'][project_id]
  project_result['project_info'] = project
  # Iterate over discovered service accounts by attempting impersonation
  project_result['service_account_edges'] = []
  updated_chain = chain_so_far + [sa_name]

  if scan_config is not None:
    impers = scan_config.get('service_accounts', None)
  else:
    impers = {'impersonate': False}  # do not impersonate by default

  # trying to impersonate SAs within project
  if impers is not None and impers.get('impersonate', False) is True:
    logging.info('Looking for impersonation options in %s', project_id)
    iam_client = iam_client_for_credentials(credentials)
    if is_set(scan_config, 'iam_policy') is False:
      iam_policy = CrawlerFactory.create_crawler('iam_policy').crawl(
          project_id,
          ClientFactory.get_client('cloudresourcemanager').get_service(
              credentials,
          ),
      )

    project_service_accounts = get_sas_for_impersonation(iam_policy)
    for candidate_service_account in project_service_accounts:
      try:
        logging.info('Trying %s', candidate_service_account)
        creds_impersonated = credsdb.impersonate_sa(
            iam_client, candidate_service_account
        )
        context.service_account_queue.put(
            (candidate_service_account, creds_impersonated, updated_chain)
        )
        project_result['service_account_edges'].append(
            candidate_service_account
        )
        logging.info(
            'Successfully impersonated %s using %s',
            candidate_service_account,
            sa_name,
        )
      except Exception:
        logging.error('Failed to get token for %s', candidate_service_account)
        logging.error(sys.exc_info()[1])


def iam_client_for_credentials(
    credentials: Credentials,
) -> IAMCredentialsClient:
  return iam_credentials.IAMCredentialsClient(credentials=credentials)


def gke_client_for_credentials(
    credentials: Credentials,
) -> container_v1.services.cluster_manager.client.ClusterManagerClient:
  return container_v1.services.cluster_manager.ClusterManagerClient(
      credentials=credentials
  )


def get_sa_details_from_key_files(key_path):
  malformed_keys = []
  sa_details = []
  for keyfile in os.listdir(key_path):
    if not keyfile.endswith('.json'):
      malformed_keys.append(keyfile)
      continue

    full_key_path = os.path.join(key_path, keyfile)
    try:
      account_name, credentials = credsdb.get_creds_from_file(full_key_path)
      if credentials is None:
        logging.error('Failed to retrieve credentials for %s', account_name)
        continue

      sa_details.append((account_name, credentials, []))
    except (MalformedError, JSONDecodeError, Exception):
      malformed_keys.append(keyfile)

  if len(malformed_keys) > 0:
    for malformed_key in malformed_keys:
      logging.error('Failed to parse keyfile: %s', malformed_key)

  return sa_details


def get_sas_for_impersonation(iam_policy: List[Dict[str, Any]]) -> List[str]:
  """Extract a list of unique SAs from IAM policy associated with project.

  Args:
    iam_policy: An IAM policy provided by get_iam_policy function.

  Returns:
    A list of service accounts represented as string
  """

  if not iam_policy:
    return []

  list_of_sas = list()
  for entry in iam_policy:
    for sa_name in entry.get('members', []):
      if sa_name.startswith('serviceAccount') and '@' in sa_name:
        account_name = sa_name.split(':')[1]
        if account_name not in list_of_sas:
          list_of_sas.append(account_name)

  return list_of_sas


def infinite_defaultdict():
  """Initialize infinite default.

  Returns:
    DefaultDict
  """
  return collections.defaultdict(infinite_defaultdict)


def get_sa_tuples(args):
  """The function extracts service account (SA) credentials from various

  sources and returns a list of tuples.
  """

  sa_tuples = []
  if args.key_path:
    # extracting SA keys from folder
    sa_tuples.extend(get_sa_details_from_key_files(args.key_path))

  if args.use_metadata:
    # extracting GCP credentials from instance metadata
    account_name, credentials = credsdb.get_creds_from_metadata()
    if credentials is None:
      logging.error('Failed to retrieve credentials from metadata')
    else:
      sa_tuples.append((account_name, credentials, []))

  if args.gcloud_profile_path:
    # extracting GCP credentials from gcloud configs
    auths_list = credsdb.get_account_creds_list(args.gcloud_profile_path)

    for accounts in auths_list:
      for creds in accounts:
        # switch between accounts
        account_name = creds.account_name
        account_creds = creds.creds
        access_token = creds.token

        if args.key_name and args.key_name not in account_name:
          continue

        logging.info('Retrieving credentials for %s', account_name)
        credentials = credsdb.get_creds_from_data(
            access_token, json.loads(account_creds)
        )
        if credentials is None:
          logging.error('Failed to retrieve access token for %s', account_name)
          continue

        sa_tuples.append((account_name, credentials, []))

  if args.access_token_files:
    for access_token_file in args.access_token_files.split(','):
      credentials = credsdb.creds_from_access_token(access_token_file)

      if credentials is None:
        logging.error('Failed to retrieve credentials using token provided')
      else:
        token_file_name = os.path.basename(access_token_file)
        sa_tuples.append((token_file_name, credentials, []))

  if args.refresh_token_files:
    for refresh_token_file in args.refresh_token_files.split(','):
      credentials = credsdb.creds_from_refresh_token(refresh_token_file)

      if credentials is None:
        logging.error('Failed to retrieve credentials using token provided')
      else:
        token_file_name = os.path.basename(refresh_token_file)
        sa_tuples.append((token_file_name, credentials, []))

  return sa_tuples


def main():
  """The main scanner loop for GCP Scanner"""

  logging.getLogger('googleapiclient.discovery_cache').setLevel(logging.ERROR)
  logging.getLogger('googleapiclient.http').setLevel(logging.ERROR)

  args = arguments.arg_parser()

  logging.basicConfig(
      level=getattr(logging, args.log_level.upper(), None),
      format='%(asctime)s - %(levelname)s - %(message)s',
      datefmt='%Y-%m-%d %H:%M:%S',
      filename=args.log_file,
      filemode='a',
  )

  force_projects_list = list()
  if args.force_projects:
    force_projects_list = args.force_projects.split(',')

  sa_tuples = scanner.get_sa_tuples(args)

  scan_config = None
  if args.config_path is not None:
    with open(args.config_path, 'r', encoding='utf-8') as f:
      scan_config = json.load(f)

  # Generate current timestamp to append to the filename
  scan_time_suffix = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')

  context = models.SpiderContext(sa_tuples)

  project_queue = list()
  processed_sas = set()

  while not context.service_account_queue.empty():
    # Get a new candidate service account / token
    sa_name, credentials, chain_so_far = context.service_account_queue.get()
    if sa_name in processed_sas:
      continue

    # Don't process this service account again
    processed_sas.add(sa_name)

    logging.info('>> current service account: %s', sa_name)
    sa_results = scanner.infinite_defaultdict()
    # Log the chain we used to get here (even if we have no privs)
    sa_results['service_account_chain'] = chain_so_far
    sa_results['current_service_account'] = sa_name
    # Add token scopes in the result
    sa_results['token_scopes'] = credentials.scopes

    project_list = CrawlerFactory.create_crawler(
        'project_list',
    ).crawl(
        ClientFactory.get_client('cloudresourcemanager').get_service(
            credentials
        ),
    )

    if len(project_list) <= 0:
      logging.info('Unable to list projects accessible from service account')

    if force_projects_list:
      for force_project_id in force_projects_list:
        if (
            next(
                (
                    item
                    for item in project_list
                    if item['projectId'] == force_project_id
                ),
                None,
            )
            is not None
        ):
          logging.info('The project %s is already in the list',
                       force_project_id)
          continue
        res = CrawlerFactory.create_crawler(
            'project_info',
        ).crawl(
            force_project_id,
            ClientFactory.get_client('cloudresourcemanager').get_service(
                credentials,
            ),
        )
        if res:
          project_list.append(res)
        else:
          # force object creation anyway
          project_list.append(
              {'projectId': force_project_id, 'projectNumber': 'N/A'}
          )

    # Enumerate projects accessible by SA
    for project in project_list:
      project_obj = models.ProjectInfo(
          project,
          sa_results,
          args.output,
          scan_config,
          args.light_scan,
          args.target_project,
          scan_time_suffix,
          sa_name,
          credentials,
          chain_so_far,
          int(args.resource_worker_count),
      )
      project_queue.append(project_obj)
      impersonate_service_accounts(
          context,
          project,
          scan_config,
          sa_results,
          chain_so_far,
          sa_name,
          credentials,
      )

  all_thread_handles = list()

  # See i#267 on why we use the native threading approach here.
  for i, project_obj in enumerate(project_queue):
    logging.info('Finished %d projects out of %d', i, len(project_queue) - 1)
    sync_t = threading.Thread(target=scanner.get_resources, args=(project_obj,))
    sync_t.daemon = True
    sync_t.start()
    all_thread_handles.append(sync_t)

    while True:  # enforce explicit block on number of threads
      active_threads = 0
      for t in all_thread_handles:
        if t.is_alive():
          active_threads += 1

      if active_threads >= int(args.project_worker_count):
        time.sleep(0.1)
      else:
        break

  # wait for any threads left to finish
  for t in all_thread_handles:
    t.join()

  return 0



#  Copyright 2023 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
This module provides miscellaneous crawler methods to crawl GCP resources using REST API
"""
import logging
import sys
from typing import List, Dict, Any, Tuple

import requests
from google.cloud import container_v1
from requests.auth import HTTPBasicAuth


def get_gke_clusters(
  project_name: str, gke_client: container_v1.services.cluster_manager.client
  .ClusterManagerClient
) -> List[Tuple[str, str]]:
  """Retrieve a list of GKE clusters available in the project.

  Args:
    project_name: A name of a project to query info about.
    gke_client: I do not know TBD.

  Returns:
    A list of GKE clusters in the project.
  """

  logging.info("Retrieving list of GKE clusters")
  parent = f"projects/{project_name}/locations/-"
  try:
    clusters = gke_client.list_clusters(parent=parent)
    return [(cluster.name, cluster.description)
            for cluster in clusters.clusters]
  except Exception:
    logging.info("Failed to retrieve cluster list for project %s", project_name)
    logging.info(sys.exc_info())
    return []


def get_gke_images(project_name: str, access_token: str) -> Dict[str, Any]:
  """Retrieve a list of GKE images available in the project.

  Args:
    project_name: A name of a project to query info about.
    access_token: An Oauth2 token with permissions to query list of gke images.

  Returns:
    A gke images JSON object for each accessible zone.
  """

  images = dict()
  logging.info("Retrieving list of GKE images")
  project_name = project_name.replace(":", "/")
  regions = ["", "us.", "eu.", "asia."]
  for region in regions:
    gcr_url = f"https://{region}gcr.io/v2/{project_name}/tags/list"
    try:
      res = requests.get(
        gcr_url, auth=HTTPBasicAuth("oauth2accesstoken", access_token),
        timeout=120)
      if not res.ok:
        logging.info("Failed to retrieve gcr images list. Status code: %d",
                     res.status_code)
        continue
      images[region.replace(".", "")] = res.json()
    except Exception:
      logging.info("Failed to retrieve gke images for project %s", project_name)
      logging.info(sys.exc_info())

  return images



#  Copyright 2023 Google LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import sys
from typing import List, Dict, Any, Union

from googleapiclient import discovery

from gcp_scanner.crawler.interface_crawler import ICrawler


class BigQueryCrawler(ICrawler):
  '''Handle crawling of BigQuery data.'''

  def crawl(self, project_id: str, service: discovery.Resource,
            config: Dict[str, Union[bool, str]] = None) -> Dict[str, List[Dict[str, Any]]]:
    '''Retrieve a list of BigQuery datasets available in the project.

    Args:
      project_id: A name of a project to query info about.
      service: A resource object for interacting with the BigQuery API.
      config: Configuration options for the crawler (Optional).

    Returns:
      A list of resource objects representing the crawled data.
    '''

    logging.info("Retrieving BigQuery Datasets")
    bq_datasets = dict()
    try:
      request = service.datasets().list(projectId=project_id)
      while request is not None:
        response = request.execute()

        for dataset in response.get("datasets", []):
          dataset_id = dataset["datasetReference"]["datasetId"]
          bq_datasets[dataset_id] = self.get_bq_tables(project_id, dataset_id, service)

        request = service.datasets().list_next(
          previous_request=request, previous_response=response)
    except Exception:
      logging.info("Failed to retrieve BQ datasets for project %s", project_id)
      logging.info(sys.exc_info())
    return bq_datasets

  def get_bq_tables(self, project_id: str, dataset_id: str,
                    bq_service: discovery.Resource) -> List[Dict[str, Any]]:
    """Retrieve a list of BigQuery tables available in the dataset.

    Args:
      project_id: A name of a project to query info about.
      dataset_id: A name of dataset to query data from.
      bq_service: I do not know.

    Returns:
      A list of BigQuery tables in the dataset.
    """

    logging.info("Retrieving BigQuery Tables for dataset %s", dataset_id)
    list_of_tables = list()
    try:
      request = bq_service.tables().list(
        projectId=project_id, datasetId=dataset_id)
      while request is not None:
        response = request.execute()
        list_of_tables.extend(response.get("tables", []))
        request = bq_service.tables().list_next(
          previous_request=request, previous_response=response)
    except Exception:
      logging.info("Failed to retrieve BQ tables for dataset %s", dataset_id)
      logging.info(sys.exc_info())
    return list_of_tables



#  Copyright 2023 Google LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import sys
from typing import List, Dict, Any

from googleapiclient import discovery

from gcp_scanner.crawler.interface_crawler import ICrawler


class CloudResourceManagerProjectListCrawler(ICrawler):
  '''Handle crawling of Cloud Resource Manager Project List data.'''

  def crawl(self, service: discovery.Resource) -> List[Dict[str, Any]]:
    '''Retrieve a list of projects accessible by credentials provided.

    Args:
      service: A resource object for interacting with the Cloud Source API.

    Returns:
      A list of resource objects representing the crawled data.
    '''

    logging.info("Retrieving projects list")
    project_list = list()
    try:
      request = service.projects().list()
      while request is not None:
        response = request.execute()
        project_list.extend(response.get("projects",[]))
        request = service.projects().list_next(
            previous_request=request, previous_response=response)
    except Exception:
      logging.info("Failed to enumerate projects")
      logging.info(sys.exc_info())
    return project_list



#  Copyright 2023 Google LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import sys
from typing import Dict, Any, Union

from googleapiclient import discovery

from gcp_scanner.crawler.interface_crawler import ICrawler


class SpannerInstancesCrawler(ICrawler):
  '''Handle crawling of Spanner Instances data.'''

  def crawl(self, project_id: str, service: discovery.Resource,
            config: Dict[str, Union[bool, str]] = None) -> Dict[str, Any]:
    '''Retrieve a list of Spanner instances available in the project.

    Args:
      project_id: A name of a project to query info about.
      service: A resource object for interacting with the Spanner API.
      config: Configuration options for the crawler (Optional).

    Returns:
      A list of resource objects representing the crawled data.
    '''

    logging.info("Retrieving spanner instances")
    spanner_instances_list = list()
    try:
      request = service.projects().instances().list(
        parent=f"projects/{project_id}")
      while request is not None:
        response = request.execute()
        spanner_instances_list.extend(response.get("instances", []))
        request = service.projects().instances().list_next(
          previous_request=request, previous_response=response)
    except Exception:
      logging.info("Failed to retrieve Spanner instances for project %s",
                   project_id)
      logging.info(sys.exc_info())
    return spanner_instances_list



#  Copyright 2023 Google LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
import sys
from typing import List, Dict, Any, Union

from googleapiclient import discovery

from gcp_scanner.crawler.interface_crawler import ICrawler


class CloudResourceManagerIAMPolicyCrawler(ICrawler):
  '''Handle crawling of Cloud Resource Manager IAM Policy data.'''

  def crawl(self, project_name: str, service: discovery.Resource,
            config: Dict[str, Union[bool, str]] = None) -> List[Dict[str, Any]]:
    '''Retrieve an IAM Policy in the project.

    Args:
      project_name: A name of a project to query info about.
      service: A resource object for interacting with the cloud source API.
      config: Configuration options for the crawler (Optional).

    Returns:
      A list of resource objects representing the crawled data.
    '''

    logging.info("Retrieving IAM policy for %s", project_name)

    resource = project_name

    get_policy_options = {"options": {"requestedPolicyVersion": 3}}
    try:
      request = service.projects().getIamPolicy(
        resource=resource, body=get_policy_options)
      response = request.execute()
    except Exception:
      logging.info("Failed to get endpoints list for project %s", project_name)
      logging.info(sys.exc_info())
      return None

    if response.get("bindings", None) is not None:
      return response["bindings"]
    else:
      return None



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""The module to handle GCP credentials.

"""

import collections
import datetime
import json
import logging
import os
import sqlite3
import sys
from typing import List, Dict, Tuple, Optional, Mapping, Union

from google.cloud.iam_credentials_v1.services.iam_credentials.client import IAMCredentialsClient
from google.oauth2 import credentials
from google.oauth2 import service_account
from httplib2 import Credentials
import requests

credentials_db_search_places = ["/home/", "/root/"]


def credentials_from_token(access_token: str, refresh_token: Optional[str],
                           token_uri: Optional[str], client_id: Optional[str],
                           client_secret: Optional[str],
                           scopes_user: Optional[str]) -> Credentials:
  return credentials.Credentials(
    access_token,
    refresh_token=refresh_token,
    token_uri=token_uri,
    client_id=client_id,
    client_secret=client_secret,
    scopes=scopes_user)


def get_creds_from_file(file_path: str) -> Tuple[str, Credentials]:
  """Creates a Credentials instance from a service account json file.

  Args:
    file_path: The path to the service account json file.

  Returns:
    str: An email address associated with a service account.
    google.auth.service_account.Credentials: The constructed credentials.
  """

  logging.info("Retrieving credentials from %s", file_path)
  creds = service_account.Credentials.from_service_account_file(file_path)
  return creds.service_account_email, creds


def get_creds_from_json(parsed_keyfile: Mapping[str, str]) -> Credentials:
  """Creates a Credentials instance from parsed service account info..

  Args:
    parsed_keyfile: The service account info in Google format.

  Returns:
    google.auth.service_account.Credentials: The constructed credentials.
  """

  return service_account.Credentials.from_service_account_info(parsed_keyfile)


def get_creds_from_metadata() -> Tuple[Optional[str], Optional[Credentials]]:
  """Retrieves a Credentials instance from compute instance metadata.

  Returns:
    str: An email associated with credentials.
    google.auth.service_account.Credentials: The constructed credentials.
  """

  print("Retrieving access token from instance metadata")

  token_url = "http://metadata.google.internal/computeMetadata/v1/instance/\
service-accounts/default/token"
  scope_url = "http://metadata.google.internal/computeMetadata/v1/instance/\
service-accounts/default/scopes"
  email_url = "http://metadata.google.internal/computeMetadata/v1/instance/\
service-accounts/default/email"
  headers = {"Metadata-Flavor": "Google"}
  try:
    res = requests.get(token_url, headers=headers, timeout=120)
    if not res.ok:
      logging.error("Failed to retrieve instance token. Status code %d",
                    res.status_code)
      return None, None
    token = res.json()["access_token"]

    res = requests.get(scope_url, headers=headers, timeout=120)
    if not res.ok:
      logging.error("Failed to retrieve instance scopes. Status code %d",
                    res.status_code)
      return None, None
    instance_scopes = res.content.decode("utf-8")

    res = requests.get(email_url, headers=headers, timeout=120)
    if not res.ok:
      logging.error("Failed to retrieve instance email. Status code %d",
                    res.status_code)
      return None, None
    email = res.content.decode("utf-8")

  except Exception:
    logging.error("Failed to retrieve instance metadata")
    logging.error(sys.exc_info()[1])
    return None, None

  print("Successfully retrieved instance metadata")
  logging.info("Access token length: %d", len(token))
  logging.info("Instance email: %s", email)
  logging.info("Instance scopes: %s", instance_scopes)
  return email, credentials_from_token(token, None, None, None, None,
                                       instance_scopes)


def get_creds_from_data(access_token: str,
                        parsed_keyfile: Dict[str, str]) -> Credentials:
  """Creates a Credentials instance from parsed service account info.

  The function currently supports two types of credentials.  Service account key
  in json format and user account with refresh token.

  Args:
    access_token: An Oauth2 access token. It can be None.
    parsed_keyfile: The service account info in Google format.

  Returns:
    google.auth.service_account.Credentials: The constructed credentials.
  """
  creds = None
  if "refresh_token" in parsed_keyfile:
    logging.info("Identified user credentials in gcloud profile")
    # this is user account credentials with refresh token
    creds = credentials_from_token(access_token,
                                   parsed_keyfile["refresh_token"],
                                   parsed_keyfile["token_uri"],
                                   parsed_keyfile["client_id"],
                                   parsed_keyfile["client_secret"],
                                   parsed_keyfile["scopes"])
  elif "private_key" in parsed_keyfile:
    logging.info("Identified service account key credentials in gcloud profile")
    # this is a service account key with private key
    creds = get_creds_from_json(parsed_keyfile)
  else:
    logging.error("unknown type of credentials")

  return creds


def find_creds(explicit_path: Optional[str] = None) -> List[str]:
  """The function search disk and returns a list of files with GCP credentials.

  Args:
    explicit_path: An explicit path on disk to search. If None, the function
      searches in standard locations where gcloud profiles are usually located.

  Returns:
    list: The list of files with GCP credentials.
  """

  logging.info("Searching for credentials on disk")
  list_of_creds_files = list()
  search_paths = list()
  if explicit_path is not None and explicit_path != "-":
    search_paths.append(explicit_path)
  else:
    credentials_db_search_places.append(os.getenv("HOME") + "/")
    for dir_path in credentials_db_search_places:
      if not os.access(dir_path, os.R_OK):
        continue
      for subdir_name in os.listdir(dir_path):
        full_path = dir_path + subdir_name + "/gcloud/"
        search_paths.append(full_path)

  for dir_path in search_paths:
    logging.info("Scanning %s for credentials.db", dir_path)
    full_path = os.path.join(dir_path, "credentials.db")
    if os.path.exists(full_path) and os.access(full_path, os.R_OK):
      print(f"Identified accessible gcloud config profile {full_path}")
      list_of_creds_files.append(full_path)
  print(f"Identified {len(list_of_creds_files)} credential DBs")
  return list_of_creds_files


def get_access_tokens_dict(path_to_creds_db: str) -> Dict[str, str]:
  """The function search and extract Oauth2 access_tokens from sqlite3 DB.

  Args:
    path_to_creds_db: A path to sqllite3 DB with gcloud access tokens.

  Returns:
    dict: The dictionary of account names and corresponding tokens.
  """

  access_tokens_dict = dict()
  access_tokens_path = path_to_creds_db.replace("credentials.db",
                                                "access_tokens.db")
  if os.path.exists(access_tokens_path) and os.access(access_tokens_path,
                                                      os.R_OK):
    logging.info("Identified access tokens DB in %s", access_tokens_path)
    conn = sqlite3.connect(access_tokens_path)
    cursor = conn.execute(
      "SELECT account_id, access_token, token_expiry FROM access_tokens")
    rows = cursor.fetchall()
    for row in rows:
      associated_account = row[0]
      token = row[1]
      expiration_date = row[2]
      expiration_date = expiration_date.split(".")[0]  # omit milliseconds

      token_time_obj = datetime.datetime.strptime(expiration_date,
                                                  "%Y-%m-%d %H:%M:%S")
      if datetime.datetime.now() > token_time_obj:
        logging.info("Token for %s expired", associated_account)
        continue

      access_tokens_dict[associated_account] = token

  return access_tokens_dict


def extract_creds(path_to_creds_db: str) -> List[Tuple[str, str, str]]:
  """The function extract refresh and associated access tokens from sqlite3 DBs.

  Args:
    path_to_creds_db: A path to sqllite3 DB with gcloud refresh tokens.

  Returns:
    list of tuples: (account name, refresh token, access token).
  """

  logging.info("Opening %s DB", path_to_creds_db)
  SA = collections.namedtuple("SA", "account_name, creds, token")

  res = list()
  conn = sqlite3.connect(path_to_creds_db)
  cursor = conn.execute("SELECT account_id, value FROM credentials")
  rows = cursor.fetchall()
  if len(rows) <= 0:
    logging.error("Empty database")
    return None
  # we also want to check for access_tokens to avoid unnecessary refreshing
  access_tokens = get_access_tokens_dict(path_to_creds_db)
  for row in rows:
    access_token = None
    if access_tokens.get(row[0], None) is not None:
      logging.info("Found valid access token for %s", row[0])
      access_token = access_tokens[row[0]]
    res.append(SA(row[0], row[1], access_token))
  print(f"Identified {len(res)} credential entries")
  return res


def get_account_creds_list(
    gcloud_profile_path: Optional[str] = None
) -> List[List[Tuple[str, str, str]]]:
  """The function searches and extracts gcloud credentials from disk.

  Args:
    gcloud_profile_path: An explicit gcloud profile path on disk to search. If
      None, the function searches in standard locations where gcloud profiles
      are usually located.

  Returns:
    list: A list of tuples (account name, refresh token, access token).
  """

  accounts = list()
  creds_file_list = find_creds(gcloud_profile_path)
  for creds_file in creds_file_list:
    res = extract_creds(creds_file)
    if res is not None:
      accounts.append(res)
  return accounts


def impersonate_sa(iam_client: IAMCredentialsClient,
                   target_account: str) -> Credentials:
  """The function is used to impersonate SA.

  Args:
    iam_client: google.cloud.iam_credentials_v1.services.iam_credentials.
      client.IAMCredentialsClient object.
    target_account: Name of a service account to impersonate.

  Returns:
    google.auth.service_account.Credentials: The constructed credentials.
  """

  scopes_sa = ["https://www.googleapis.com/auth/cloud-platform"]
  intermediate_access_token = iam_client.generate_access_token(
    name=target_account, scope=scopes_sa, retry=None
    # lifetime = "43200"
  )

  return credentials_from_token(intermediate_access_token.access_token, None,
                                None, None, None, scopes_sa)


def creds_from_access_token(access_token_file):
  """The function is used to obtain Google Auth Credentials from access token.

  Args:
    access_token_file: a path to a file with access token and scopes stored in
    JSON format. Example:
      {
        "access_token": "<token>",
        "scopes": [
          "https://www.googleapis.com/auth/devstorage.read_only",
          "https://www.googleapis.com/auth/logging.write",
          "https://www.googleapis.com/auth/monitoring.write",
          "https://www.googleapis.com/auth/servicecontrol",
          "https://www.googleapis.com/auth/service.management.readonly",
          "https://www.googleapis.com/auth/trace.append"
        ]
      }

  Returns:
    google.auth.service_account.Credentials: The constructed credentials.
  """

  with open(access_token_file, encoding="utf-8") as f:
    creds_dict = json.load(f)

  user_scopes = creds_dict.get("scopes", None)
  if user_scopes is None:
    user_scopes = ["https://www.googleapis.com/auth/cloud-platform"]

  return credentials_from_token(
    creds_dict["access_token"],
    None,
    None,
    None,
    None,
    user_scopes)


def creds_from_refresh_token(refresh_token_file):
  """The function is used to obtain Google Auth Credentials from refresh token.

  Args:
    refresh_token_file: a path to a file with refresh_token, client_id,
      client_secret, and token_uri stored in JSON format.
    Example:
      {
        "refresh_token": "<token>",
        "client_id": "id",
        "client_secret": "secret",
        scopes: [
          https://www.googleapis.com/auth/devstorage.read_only,
          https://www.googleapis.com/auth/logging.write,
          https://www.googleapis.com/auth/monitoring.write,
          https://www.googleapis.com/auth/servicecontrol,
          https://www.googleapis.com/auth/service.management.readonly,
          https://www.googleapis.com/auth/trace.append
        ]
      }
  Returns:
    google.auth.service_account.Credentials: The constructed credentials.
  """

  with open(refresh_token_file, encoding="utf-8") as f:
    creds_dict = json.load(f)

  user_scopes = get_scopes_from_refresh_token(creds_dict)

  return credentials.Credentials(
    None,
    refresh_token=creds_dict["refresh_token"],
    token_uri=creds_dict["token_uri"],
    client_id=creds_dict["client_id"],
    client_secret=creds_dict["client_secret"],
    scopes=user_scopes,
  )


def get_scopes_from_refresh_token(context) -> Union[List[str], None]:
  """The function is used to obtain scopes from refresh token.

  Args:
    context: dictionary containing refresh_token data
    Example:
      {
        "refresh_token": "<token>",
        "client_id": "id",
        "client_secret": "secret",
      }
  Returns:
    a list of scopes or None
  """
  # Obtain access token from the refresh token
  token_uri = "https://oauth2.googleapis.com/token"
  context["grant_type"] = "refresh_token"

  try:
    response = requests.post(token_uri, data=context, timeout=5)
    # prepare the scope string into a list
    raw = response.json().get("scope", None)
    return raw.split(" ") if raw else None
  except Exception as ex:
    logging.error(
      "Failed to retrieve access token from refresh token.",
    )
    logging.debug("Token refresh exception", exc_info=ex)

  return None



